from collections.abc import Callable, Iterable, Sequence

import ida_hexrays
import idaapi
from ida_funcs import func_t
from ida_hexrays import cexpr_t, ctree_parentee_t, lvar_t
from ida_typeinf import tinfo_t
from idahelper import tif
from idahelper.ast import cfunc
from idahelper.ast.lvars import VariableModification
from idahelper.pac import client as pac

from ioshelper.base.utils import CustomDict, CustomSet

from .renamer import (
    Modifications,
)
from .visitor import Call, XrefsMatcher, process_function_calls


def apply_pac(func: func_t) -> bool:
    print(f"Trying to apply pac signature on current function: {func.start_ea:X}")
    helper = MostSpecificAncestorHelper()
    xref_matcher = XrefsMatcher.build([], on_unknown_call_wrapper(helper))  # type: ignore  # noqa: PGH003
    decompiled_func = cfunc.from_func(func)
    if decompiled_func is None:
        return False

    with Modifications(decompiled_func.entry_ea, decompiled_func.get_lvars()) as modifications:
        process_function_calls(decompiled_func.mba, xref_matcher, modifications)
        has_changed = False
        for lvar, typ, calls in helper.lvars():
            if not lvar.has_user_type and should_modify_type(lvar.type(), typ):
                has_changed = True
                print(f"Modifying lvar {lvar.name} from {lvar.type()}* to {typ}")
                modifications.modify_local(lvar.name, VariableModification(type=tif.pointer_of(typ)))
                fix_calls(typ, calls)

        for (cls_type, offset), typ, calls in helper.fields():
            typ_res = tif.get_member_recursive(cls_type, offset)
            if typ_res is None:
                print(f"[Warning] Could not find member at offset {offset:X} in {cls_type}")
                continue
            actual_typ, member = typ_res
            if member.size == 64 and should_modify_type(member.type, typ):
                has_changed = True
                print(f"Modifying field {cls_type}::{member.name} from {member.type}* to {typ}")
                # noinspection PyTypeChecker
                modifications.modify_type(
                    actual_typ.get_type_name(),  # pyright: ignore [reportArgumentType]
                    offset,
                    VariableModification(type=tif.pointer_of(typ)),
                )
                fix_calls(typ, calls)

    if has_changed:
        fix_empty_calls(func)

    return has_changed


def fix_empty_calls(func: func_t):
    """If after the modifications there are calls with no parameters, do force the call type as it is better then nothing"""
    decompiled_func = ida_hexrays.decompile(func.start_ea, flags=ida_hexrays.DECOMP_NO_CACHE)
    if decompiled_func is None:
        print(f"[Warning] Could not decompile function {func.start_ea:X} to fix empty calls")
        return

    # noinspection PyTypeChecker
    EmptyCallTreeVisitor().apply_to(decompiled_func.body, None)  # pyright: ignore[reportIncompatibleMethodCall]


class EmptyCallTreeVisitor(ctree_parentee_t):
    def visit_expr(self, expr: cexpr_t) -> int:  # pyright: ignore[reportIncompatibleMethodOverride]
        # Filter dynamic calls with no parameters
        if (
            expr.op != ida_hexrays.cot_call
            or not expr.x.type.is_funcptr()
            or expr.a.size() != 0
            or expr.ea == idaapi.BADADDR
        ):
            return 0

        # Make sure it is a call to a vtable member
        x = expr.x
        # Handle cast of the function type
        if x.op == ida_hexrays.cot_cast:
            x = x.x

        # Check it is a member
        if x.op != ida_hexrays.cot_memptr:
            return 0

        # Check it is a member of a vtable
        possible_vtable_type: tinfo_t = x.x.type
        if not possible_vtable_type.is_ptr() or not possible_vtable_type.get_pointed_object().is_vftable():
            return 0

        # This is a vtable call with no parameters, apply the type to the call
        apply_vtable_type_to_call(expr.ea, possible_vtable_type.get_pointed_object(), expr.x.m, apply_if_no_args=True)
        return 0


def fix_calls(class_type: tinfo_t, calls: list[Call]):
    """Apply the call type from vtable definition to each of the calls in the list."""
    if not calls:
        return

    vtable_type = tif.vtable_type_from_type(class_type)
    if vtable_type is None:
        print(f"[Warning] Could not find vtable type for {class_type}")
        return

    for call in calls:
        assert call.indirect_info is not None
        offset = call.indirect_info.offset
        apply_vtable_type_to_call(call.ea, vtable_type, offset, apply_if_no_args=False)


def apply_vtable_type_to_call(call_ea: int, vtable_type: tinfo_t, offset: int, apply_if_no_args: bool) -> bool:
    """Apply the vtable type to the call at the given ea."""
    vtable_member = tif.get_member(vtable_type, offset)
    if vtable_member is None:
        print(f"[Warning] Could not find vtable member for {vtable_type} at {offset:X}")
        return False

    # There are a lot of false positive signatures that have only "this" argument.
    # We prefer not to force non-arguments calls rather than hide arguments.
    vtable_member_type: tinfo_t = tinfo_t(vtable_member.type)
    vtable_member_type.remove_ptr_or_array()
    if apply_if_no_args or vtable_member_type.get_nargs() != 1:
        tif.apply_tinfo_to_call(vtable_member.type, call_ea)
        print(
            f"Applying vtable type {vtable_member.type} to call at {call_ea:X} for {vtable_type}::{vtable_member.name} at offset {offset:X}"
        )
        return True
    return False


def should_modify_type(current_type: tinfo_t, new_type: tinfo_t) -> bool:
    if current_type == new_type:
        return False
    elif not current_type.is_ptr():
        return True
    current_type.remove_ptr_or_array()

    if current_type.is_void() or not current_type.is_struct() or current_type.get_type_name() == "OSObject":
        return True

    return current_type in tif.get_parent_classes(new_type, True)


class MostSpecificAncestorHelper:
    """Helper class to find the most specific ancestor of a PAC call."""

    def __init__(self):
        self._lvars: CustomDict[lvar_t, CustomSet[tinfo_t]] = CustomDict(lambda v: v.name)
        self._fields: CustomDict[tuple[tinfo_t, int], CustomSet[tinfo_t]] = CustomDict(lambda t: (t[0].get_tid(), t[1]))

        self._lvar_to_calls: CustomDict[lvar_t, list[Call]] = CustomDict(lambda v: v.name)
        self._fields_to_calls: CustomDict[tuple[tinfo_t, int], list[Call]] = CustomDict(
            lambda t: (t[0].get_tid(), t[1])
        )

        self._children_classes_cache: CustomDict[tinfo_t, list[tinfo_t]] = CustomDict(lambda t: t.get_tid())
        """Cache of mapping from a type to its children classes."""

    def update_lvar(self, lvar: lvar_t, predicate: Sequence[tinfo_t], call: Call):
        minimized = self._minimize(predicate)
        if lvar in self._lvars:
            self._lvars[lvar] &= self._children_of_union(minimized)
        else:
            self._lvars[lvar] = self._children_of_union(minimized)

        self._lvar_to_calls.setdefault(lvar, []).append(call)

    def update_field(self, cls_type: tinfo_t, offset: int, predicate: Sequence[tinfo_t], call: Call):
        if cls_type.is_vftable():
            return

        minimized = self._minimize(predicate)
        key = (cls_type, offset)
        if key in self._fields:
            self._fields[key] &= self._children_of_union(minimized)
        else:
            self._fields[key] = self._children_of_union(minimized)

        self._fields_to_calls.setdefault(key, []).append(call)

    def lvars(self) -> Iterable[tuple[lvar_t, tinfo_t, list[Call]]]:
        """Get all lvars with their most specific type."""
        for lvar, state in self._lvars.items():
            ancestor = tif.get_common_ancestor(list(state))
            if ancestor is None:
                continue
            yield lvar, ancestor, self._lvar_to_calls[lvar]

    def fields(self) -> Iterable[tuple[tuple[tinfo_t, int], tinfo_t, list[Call]]]:
        """Get all fields with their most specific type."""
        for field, state in self._fields.items():
            ancestor = tif.get_common_ancestor(list(state))
            if ancestor is None:
                continue
            yield field, ancestor, self._fields_to_calls[field]

    def _get_children_classes(self, typ: tinfo_t) -> list[tinfo_t]:
        """Get all children of a type, caching the result."""
        if typ not in self._children_classes_cache:
            children = tif.get_children_classes(typ) or []
            children.append(typ)
            self._children_classes_cache[typ] = children
        return self._children_classes_cache[typ]

    def _children_of_union(self, union: tuple[tinfo_t, ...]) -> CustomSet[tinfo_t]:
        """Get all children of a union type."""
        children: CustomSet[tinfo_t] = CustomSet(lambda t: t.get_tid())
        children.add_all(union)

        for typ in union:
            children.add_all(self._get_children_classes(typ))
        return children

    @staticmethod
    def _minimize(union: Sequence[tinfo_t]) -> tuple[tinfo_t, ...]:
        """Minimize the union to its most specific types."""
        if len(union) == 1:
            return (union[0],)

        minimized: list[tinfo_t] = []

        for i, typ in enumerate(union):
            # Quick exit if the type is OSObject, as it is a base class for all objects
            if typ.get_type_name() == "OSObject":
                return (typ,)

            typ_parents = tif.get_parent_classes(typ, True)
            for j, other_typ in enumerate(union):
                if i == j:
                    continue

                if other_typ in typ_parents:
                    break
            else:
                minimized.append(typ)

        return tuple(minimized)


def on_unknown_call_wrapper(helper: MostSpecificAncestorHelper) -> Callable[[Call, Modifications], None]:
    def on_unknown_call(call: Call, _modifications: Modifications):
        """Called when a call is found"""
        if call.indirect_info is None:
            return

        prev_movk = pac.get_previous_movk(call.ea)
        if prev_movk is None:
            return
        candidates = pac.pac_class_candidates_from_movk(prev_movk)
        if candidates:
            if call.indirect_info.var is not None:
                lvar = call.indirect_info.var
                helper.update_lvar(lvar, candidates, call)
            elif call.indirect_info.field is not None:
                cls_type, offset = call.indirect_info.field
                helper.update_field(cls_type, offset, candidates, call)

    return on_unknown_call



__all__ = ["StructFieldAssignment", "get_struct_fields_assignments", "run_objc_plugin_on_func"]

from dataclasses import dataclass

import ida_hexrays
import idaapi
from ida_hexrays import cexpr_t, cfuncptr_t, cinsn_t, lvar_t
from ida_typeinf import tinfo_t, udm_t
from idahelper import tif


@dataclass
class StructFieldAssignment:
    type: tinfo_t
    member: udm_t
    expr: cexpr_t
    is_cast_assign: bool
    insn: cinsn_t

    def __repr__(self):
        return f"StructAssignment(member={self.member.name}, is_cast_assign={self.is_cast_assign},  expr={self.expr.dstr()}, insn={self.insn.dstr()})"


class LvarFieldsAssignmentsCollector(ida_hexrays.ctree_visitor_t):
    """
    Visitor for collecting assignments to a list of local variables.
    The local variables should be of struct type.
    The assignments we are searching for are for a member of the struct.

    For example: `a.b = c;` where `a` is a local variable of struct type and `b` is a member of that struct.
    """

    def __init__(self, target_lvars: list[lvar_t]):
        super().__init__(ida_hexrays.CV_PARENTS)
        self._target_lvars_names: dict[str, lvar_t] = {lvar.name: lvar for lvar in target_lvars}
        self.assignments: dict[str, list[StructFieldAssignment]] = {}

    def visit_expr(self, exp: cexpr_t) -> int:
        # We search for "a.b = c;" or "*(_XWORD *)&a.b = c;"
        # Check if the expression is an assignment
        if exp.op != ida_hexrays.cot_asg:
            return 0

        # Check if the left side is a member reference
        target = remove_ref_cast_dref(exp.x)
        is_cast_assign = target != exp.x
        if target.op != ida_hexrays.cot_memref:
            return 0

        # Check that it is a member reference to a local variable
        target_obj = target.x
        if target_obj.op != ida_hexrays.cot_var:
            return 0

        # Check if the local variable is what we are looking for
        target_obj_lvar: lvar_t = target_obj.v.getv()
        if target_obj_lvar.name not in self._target_lvars_names:
            return 0

        lvar_type = target_obj_lvar.type()
        member = tif.get_member(lvar_type, target.m)
        if member is None:
            return 0

        # Save the assignment
        self.assignments.setdefault(target_obj_lvar.name, []).append(
            StructFieldAssignment(
                type=lvar_type, member=member, expr=exp.y, insn=self.parent_insn(), is_cast_assign=is_cast_assign
            )
        )
        return 0


def remove_ref_cast_dref(expr: cexpr_t) -> cexpr_t:
    """Remove reference, cast and dereference from the expression"""
    while expr.op in (ida_hexrays.cot_ref, ida_hexrays.cot_cast, ida_hexrays.cot_ptr):
        expr = expr.x
    return expr


def get_struct_fields_assignments(cfunc: cfuncptr_t, lvars: list[lvar_t]) -> dict[str, list[StructFieldAssignment]]:
    """Get all assignments of the form "a.b = c" to the given local variables"""
    collector = LvarFieldsAssignmentsCollector(lvars)
    collector.apply_to(cfunc.body, None)  # pyright: ignore[reportArgumentType]
    return collector.assignments


def run_objc_plugin_on_func(ea: int) -> None:
    """Run IDA's Objective-C>Analyze stack-allocated blocks on the function at ea."""
    n = idaapi.netnode()
    n.create("$ objc")
    n.altset(1, ea, "R")  # the address can be any address within the function
    idaapi.load_and_run_plugin("objc", 5)



__all__ = ["locate_selector_xrefs"]

import idautils
import idc
from idahelper.widgets import EAChoose

# Taken from: https://github.com/doronz88/ida-scripts/blob/main/objc_hotkeys.py
IGNORED_SECTIONS = ("__objc_const",)


def get_name_for_ea(ea: int) -> str:
    func_name = idc.get_func_name(ea)
    func_address = idc.get_name_ea_simple(func_name)
    return func_name if ea == func_address else f"{func_name}+{ea - func_address:08x}"


def locate_selector_xrefs() -> None:
    current_ea = idc.get_screen_ea()
    func_name = idc.get_func_name(current_ea)
    try:
        selector = func_name.split(" ")[1].split("]")[0]
    except IndexError:
        print("Failed to find current selector")
        return
    print(f"looking for references to: {selector}")

    items = [
        (ea.frm, get_name_for_ea(ea.frm))
        for ea in idautils.XrefsTo(idc.get_name_ea_simple(f"_objc_msgSend${selector}"))
    ]
    if items:
        for ea, name in items:
            print(f"0x{ea:08x} {name}")
        xrefs_choose = EAChoose(f"Xrefs to selector: {selector}", items, modal=True)
        xrefs_choose.show()



import dataclasses
from enum import Enum

import ida_hexrays
from ida_hexrays import mblock_t, mcallinfo_t, minsn_t, mop_t, optblock_t
from ida_typeinf import tinfo_t
from idahelper import tif
from idahelper.microcode import mblock, mcallarg, minsn, mop
from idahelper.microcode.optimizers import optblock_counter_t, optblock_counter_wrapper_t

from .os_log import LogCallInfo, LogCallParams, get_call_info_for_name, log_type_to_str


def log_func_to_tif() -> tinfo_t | None:
    """Create tif for a log function: void f(char *fmt, ...)"""
    return tif.from_func_components("void", [tif.FuncParam("char*", "fmt"), tif.FuncParam("...")])


class ScanLogState(Enum):
    HEADER = 0
    ITEM_HEADER = 1
    ITEM_VALUE = 2


@dataclasses.dataclass
class CollectLogParamsResult:
    instructions: list[minsn_t]
    call_params: list[mop_t]


# noinspection PyMethodMayBeStatic
class log_macro_optimizer_t(optblock_counter_t):
    def func(self, blk: mblock_t) -> int:
        if blk.mba.maturity < ida_hexrays.MMAT_CALLS:
            return 0

        self.optimize_log_macro(blk)

        if self.cnt:
            blk.mark_lists_dirty()
        return self.cnt

    def optimize_log_macro(self, blk: mblock_t) -> None:
        # Find log call and extract params
        if (res := self.find_log_call(blk)) is None:
            return
        call_insn, params = res

        # Collect parameters to log
        log_params_res = self.collect_log_params(blk, params)
        if log_params_res is None:
            return

        # All instructions up to the call are part of the logging macro starting from `from_index`,
        # so they can be safely removed.
        # First, convert the call insn to the helper call
        prefix = "ossignpost_" if params.is_signpost else "os_log_"
        call_insn.l.make_helper(f"{prefix}{log_type_to_str(params.log_type, params.is_signpost)}")
        self.count()

        # Then modify callinfo
        # Remove all arguments.
        fi: mcallinfo_t = call_insn.d.f
        fi.args.clear()
        # Not necessary but IDA will crash on inconsistency, and we prefer to keep it alive if there is a bug.
        fi.solid_args = 0
        self.count()

        # Add optional name string argument
        if params.name_str_ea is not None:
            new_arg = mcallarg.from_mop(mop.from_global_ref(params.name_str_ea), tif.from_c_type("char*"))
            fi.args.push_back(new_arg)
            fi.solid_args += 1
            self.count()

        # Add format string argument
        new_arg = mcallarg.from_mop(mop.from_global_ref(params.format_str_ea), tif.from_c_type("char*"))
        fi.args.push_back(new_arg)
        fi.solid_args += 1
        self.count()

        # Add params
        for param in log_params_res.call_params:
            fi.args.push_back(mcallarg.from_mop(param, tif.from_size(param.size)))
            fi.solid_args += 1
            self.count()

        # Apply final type signature. For some reason `set_type` crashes IDA, but swap works great...
        fi.get_type().swap(log_func_to_tif())  # TODO: sometimes IDA inserts incorrect casts. Fix it.
        self.count()

        # Finally, convert other instructions (that are part of the log macro) to nop
        for insn in log_params_res.instructions:
            blk.make_nop(insn)
            self.count()

    def collect_log_params(self, blk: mblock_t, params: LogCallParams) -> CollectLogParamsResult | None:  # noqa: C901
        """
        Collect the parameters of a log macro starting from `params.call_ea` and going backwards.
        Returns the index of the first instruction that is a part of the macro and the list of parameters.
        """
        base = params.stack_base_offset
        end = base + params.size
        call_params: list[mop_t] = []
        buffer_instructions: list[minsn_t] = []
        buffer_size = 0
        size_left_for_header: int | None = None
        state = ScanLogState.HEADER

        for insn in mblock.instructions(blk):
            # Stop at the call
            if insn.ea == params.call_ea:
                break

            if not self.check_insn_part_of_log_macro(insn, params.call_ea, base, end):
                continue

            buffer_instructions.append(insn)

            # Advance state
            if state == ScanLogState.HEADER:
                if insn.l.size == 2:
                    # Only the header
                    state = ScanLogState.ITEM_HEADER
                elif insn.l.size == 4:
                    # Header + item header
                    state = ScanLogState.ITEM_VALUE
                else:
                    print(f"[Error] invalid log macro header size of {hex(params.call_ea)}: {insn.dstr()}")
                    return None
            elif state == ScanLogState.ITEM_HEADER:
                if insn.d.size != 2:
                    if insn.d.size > 2:
                        print(
                            f"[Error] Unsupported log macro because header size is bigger than 2 bytes: {insn.dstr()}"
                        )
                        return None
                    elif size_left_for_header is None:
                        size_left_for_header = 2 - insn.d.size
                    else:
                        size_left_for_header -= insn.d.size
                        if size_left_for_header == 0:
                            state = ScanLogState.ITEM_VALUE
                            size_left_for_header = None
                else:
                    state = ScanLogState.ITEM_VALUE
            else:
                call_params.append(insn.l)
                state = ScanLogState.ITEM_HEADER

            buffer_size += insn.d.size

        if state == ScanLogState.HEADER:
            # Never found the beginning of the log macro
            return None
        elif state == ScanLogState.ITEM_VALUE:
            print(f"[Error] failed to parse log macro of {hex(params.call_ea)}")
            return None

        if buffer_size != params.size:
            print(
                f"[Error] log macro size mismatch of {hex(params.call_ea)}: "
                f"expected - {params.size}, found - {buffer_size}"
            )
            return None

        return CollectLogParamsResult(buffer_instructions, call_params)

    def find_log_call(self, blk: mblock_t) -> tuple[minsn_t, LogCallParams] | None:
        """Find an `os_log` call in the given block and extract the parameters from it"""
        for insn in mblock.instructions(blk):
            if insn.opcode != ida_hexrays.m_call:
                continue

            # Search for log call
            call_name = minsn.get_func_name_of_call(insn)
            if (call_info := get_call_info_for_name(call_name)) is None:
                continue

            params = self.extract_params_from_log_call(insn, call_info)
            if params is None:
                continue

            return insn, params
        return None

    def check_insn_part_of_log_macro(
        self, insn: minsn_t, call_ea: int, base: int, end: int, print_error: bool = False
    ) -> bool:
        """Check that the given `insn` is indeed part of a log macro"""
        # It is an Assignment
        if insn.opcode not in [
            ida_hexrays.m_mov,
            ida_hexrays.m_and,
            ida_hexrays.m_xds,
            ida_hexrays.m_xdu,
            ida_hexrays.m_low,
        ]:
            if print_error:
                print(f"[Error] unsupported instruction in log block of {hex(call_ea)}: {insn.dstr()}")
            return False

        # To stack
        if insn.d.t != ida_hexrays.mop_S:
            if print_error:
                print(f"[Error] unsupported dest in log block of {hex(call_ea)}: {insn.dstr()}")
            return False

        # in range
        addr, size = insn.d.s.off, insn.d.size
        if not (base <= addr and addr + size <= end):
            if print_error:
                print(f"[Error] assignment not in range in log block of {hex(call_ea)}: {insn.dstr()}")
            return False

        return True

    def extract_params_from_log_call(self, insn: minsn_t, log_call_info: LogCallInfo) -> LogCallParams | None:
        """Given the log call instruction and information about the indices to extract, extract them"""
        call_info: mcallinfo_t = insn.d.f
        if call_info.args.empty():
            # Not calls args, probably too early in IDA's optimization
            return None

        # Get operands
        size_param = call_info.args[log_call_info.size_index]
        buf_param = call_info.args[log_call_info.buf_index]
        format_param = call_info.args[log_call_info.format_index]
        type_param = call_info.args[log_call_info.type_index]
        name_param = call_info.args[log_call_info.name_index] if log_call_info.name_index is not None else None
        # Verify types of operands
        if (
            size_param.t != ida_hexrays.mop_n
            or type_param.t != ida_hexrays.mop_n
            or not format_param.is_glbaddr()
            or buf_param.t != ida_hexrays.mop_a
            or (name_param is not None and not name_param.is_glbaddr())
        ):
            return None

        size = size_param.unsigned_value()
        log_type = type_param.unsigned_value()
        format_str_ea = format_param.a.g
        name_str_ea = name_param.a.g if name_param is not None else None
        # Check if this is a stack variable
        if buf_param.a.t != ida_hexrays.mop_S:
            return None
        stack_base_offset = buf_param.a.s.off
        return LogCallParams(
            log_type, size, stack_base_offset, format_str_ea, insn.ea, name_str_ea, log_call_info.is_signpost
        )


def optimizer() -> optblock_t:
    return optblock_counter_wrapper_t(log_macro_optimizer_t)



__all__ = ["objc_xrefs_component"]

import ida_kernwin
import idaapi

from ioshelper.base.reloadable_plugin import UIAction, UIActionsComponent

from .objc_ref import locate_selector_xrefs

ACTION_ID = "ioshelper:show_objc_xrefs"

objc_xrefs_component = UIActionsComponent.factory(
    "Show objc xrefs",
    [
        lambda core: UIAction(
            ACTION_ID,
            idaapi.action_desc_t(
                ACTION_ID, "Show xrefs for current Obj-C method's selector", ShowObjcXrefsActionHandler(), "Ctrl+4"
            ),
        )
    ],
)


class ShowObjcXrefsActionHandler(ida_kernwin.action_handler_t):
    def activate(self, ctx):
        locate_selector_xrefs()
        return 0

    def update(self, ctx) -> int:
        return idaapi.AST_ENABLE_ALWAYS



__all__ = [
    "apply_global_rename",
    "apply_pac",
    "apply_pac_component",
    "local_func_renamer_component",
    "mass_func_renamer_component",
]

import ida_kernwin
import idaapi
from ida_kernwin import action_handler_t
from idahelper import widgets

from ioshelper.base.reloadable_plugin import HexraysHookComponent, UIAction, UIActionsComponent

from .func_renamers import apply_global_rename, hooks
from .pac_applier import apply_pac

ACTION_ID = "ioshelper:func_renamer"

local_func_renamer_component = HexraysHookComponent.factory("Local rename based on function calls", [hooks])

mass_func_renamer_component = UIActionsComponent.factory(
    "Mass rename based on function calls",
    [
        lambda core: UIAction(
            ACTION_ID,
            idaapi.action_desc_t(
                ACTION_ID, "Mass rename globals and fields based on specific function calls", FuncRenameGlobalAction()
            ),
            menu_location=UIAction.base_location(core),
        )
    ],
)


class FuncRenameGlobalAction(action_handler_t):
    def activate(self, ctx: ida_kernwin.action_ctx_base_t):
        apply_global_rename()
        return 0

    def update(self, ctx):
        return idaapi.AST_ENABLE_ALWAYS


ACTION_ID_PAC = "ioshelper:apply_pac_on_function"


apply_pac_component = UIActionsComponent.factory(
    "Apply PAC types on current function",
    [
        lambda core: UIAction(
            ACTION_ID_PAC,
            idaapi.action_desc_t(ACTION_ID_PAC, "Apply PAC types on current function", ApplyPACAction(), "Ctrl+P"),
            menu_location=UIAction.base_location(core),
        )
    ],
)


class ApplyPACAction(action_handler_t):
    def activate(self, ctx: ida_kernwin.action_ctx_base_t) -> bool:
        if ctx.cur_func is None:
            print("[Error] Not inside a function")
            return False

        if apply_pac(ctx.cur_func) and ctx.widget is not None:
            widgets.refresh_widget(ctx.widget)
        return False

    def update(self, ctx):
        return idaapi.AST_ENABLE_ALWAYS



import abc
from abc import ABC
from collections.abc import Callable

import ida_hexrays
from ida_hexrays import lvars_t, minsn_t, mop_t
from ida_typeinf import tinfo_t
from idahelper import memory, tif, xrefs
from idahelper.ast import lvars
from idahelper.ast.lvars import VariableModification

from .visitor import Call, FuncXref, IndirectCallXref, ParsedParam, SourceXref


class Modifications:
    """Represents the modifications to be applied from a traversal of a single function"""

    def __init__(self, func_ea: int, func_lvars: lvars_t | None = None):
        self.func_ea = func_ea
        self.has_func_name = memory.is_user_defined_name(func_ea)
        self.lvars = func_lvars
        self._local_modifications: dict[str, VariableModification] = {}
        self._global_modifications: dict[int, VariableModification] = {}
        self._type_modifications: dict[tuple[str, int], VariableModification] = {}
        self._func_name: str | None = None

    def modify_local(self, name: str, modification: VariableModification):
        self._local_modifications[name] = self._merge_modifications(self._local_modifications.get(name), modification)

    def modify_global(self, ea: int, modification: VariableModification):
        self._global_modifications[ea] = self._merge_modifications(self._global_modifications.get(ea), modification)

    def modify_type(self, name: str, offset: int, modification: VariableModification):
        self._type_modifications[(name, offset)] = self._merge_modifications(
            self._type_modifications.get((name, offset)), modification
        )

    def set_func_name(self, name: str, force_name_change: bool = False):
        if force_name_change or not self.has_func_name:
            self._func_name = name

    def _apply(self):
        """Apply the modifications"""
        self._fix_modifications_name()
        self._apply_local_modifications()
        self._apply_global_modifications()
        self._apply_func_name()
        self._apply_type_modifications()

    def _fix_modifications_name(self):
        fix_name = lambda s: None if s is None else s.replace("-", "_").replace(" ", "_")
        for modification in self._local_modifications.values():
            modification.name = fix_name(modification.name)
        for modification in self._global_modifications.values():
            modification.name = fix_name(modification.name)

    def _apply_local_modifications(self):
        """Apply local modifications"""
        if self._local_modifications and self.lvars is not None:
            lvars.perform_lvar_modifications(self.func_ea, self.lvars, self._local_modifications, temp_fallback=True)

    def _apply_global_modifications(self):
        """Apply global modifications"""
        for ea, modification in self._global_modifications.items():
            if not modification.force_name_change and memory.is_user_defined_name(ea):
                continue
            if modification.name is not None and not memory.set_name(ea, modification.name, retry=True):
                print(f"Could not rename {hex(ea)} to {modification.name}")
            if modification.type is not None and not tif.apply_tinfo_to_ea(modification.type, ea):
                print(f"Could not retype {hex(ea)} to {modification.type}")

    def _apply_func_name(self):
        """Apply function name modification"""
        if self._func_name is not None and not memory.set_name(self.func_ea, self._func_name, retry=True):
            print(f"Could not rename {hex(self.func_ea)} to {self._func_name}")

    def _apply_type_modifications(self):
        """Apply type modifications"""
        for (name, offset), modification in self._type_modifications.items():
            cls_type = tif.from_struct_name(name)
            if cls_type is None:
                print(f"Could not find class type: {name}")
                continue
            member = tif.get_member(cls_type, offset)
            if member is None:
                print(f"Could not find member at offset: {cls_type.dstr()} at {offset}")
                return

            if modification.name is not None and not tif.set_udm_name(cls_type, member, modification.name):
                print(f"Could not rename {cls_type.dstr()} member at offset {offset} to {modification.name}")
            if modification.type is not None and not tif.set_udm_type(cls_type, member, modification.type):
                print(f"Could not retype {cls_type.dstr()} member at offset {offset} to {modification.type.dstr()}")

    # noinspection PyMethodMayBeStatic
    def _merge_modifications(
        self, original: VariableModification | None, new: VariableModification
    ) -> VariableModification:
        """Merge the modifications, preferring the new ones"""
        if original is None:
            return new

        return VariableModification(
            new.name or original.name,
            new.type or original.type,
            new.comment or original.comment,
            new.force_name_change or original.force_name_change,
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._apply()

    def __str__(self):
        return (
            f"Modifications(func_ea={self.func_ea:#x}, local_modifications={self._local_modifications}, "
            f"global_modifications={self._global_modifications}, type_modifications={self._type_modifications}, "
            f"new_name={self._func_name})"
        )


class FuncHandler(abc.ABC):
    def __init__(self, name: str):
        self.name: str = name

    @abc.abstractmethod
    def get_source_xref(self) -> SourceXref | None:
        """Get the source xref of the handler"""
        ...

    @abc.abstractmethod
    def on_call(self, call: Call, modifications: Modifications):
        """Called when a call is found"""
        ...

    def _retype_assignee(
        self, modifications: Modifications, call: Call, new_type: tinfo_t | Callable[[Call], tinfo_t | None] | None
    ):
        """Try to retype the assignee of the call to the given type"""
        if new_type is None:
            # We don't have the type, so we can't retype the assignee
            return None
        elif call.assignee is None:
            print(f"Call {call} has no assignee")
            return

        if new_type is None:
            return
        elif isinstance(new_type, tinfo_t):
            typ = new_type
        else:
            typ = new_type(call)
            if typ is None:
                return

        self.__retype_mop(call.assignee, typ, modifications)

    def _rename_parameter_by_index(
        self,
        modifications: Modifications,
        call: Call,
        name_index: int,
        rename_index: int,
        modifier: Callable[[ParsedParam], str] | None = None,
    ):
        """
        Rename the {rename_index} parameter of the call from parameter {name_index}.
        Optionally, you can pass a modifier that receives the value of the parameter and returns a new name.
        """
        if name_index >= len(call.params) or rename_index >= len(call.params):
            print(
                f"Call {call} has only {len(call.params)} parameters, tried to access {name_index} and {rename_index}"
            )
            return
        param = call.params[name_index]
        if param is None:
            print(f"Call {call} param at index {name_index} is not const: {call.params_op[name_index].dstr()}")
            return
        if modifier is None and not isinstance(param, str):
            print(f"Call {call} param at index {name_index} is not a string: {param}")
            return

        name: str = modifier(param) if modifier else param
        self.__rename_mop(call.params_op[rename_index], name, modifications)

    def _retype_parameter_by_index(
        self,
        modifications: Modifications,
        call: Call,
        index: int,
        new_type: tinfo_t | Callable[[Call], tinfo_t | None] | None,
    ):
        """
        Retype the parameter at index {index}
        """
        if index >= len(call.params):
            print(f"Call {call} has only {len(call.params)} parameters, tried to access {index}")
            return

        if new_type is None:
            return
        elif isinstance(new_type, tinfo_t):
            typ = new_type
        else:
            typ = new_type(call)
            if typ is None:
                return

        self.__retype_mop(call.params_op[index], typ, modifications)

    def _rename_assignee_by_index(
        self, modifications: Modifications, call: Call, index: int, modifier: Callable[[ParsedParam], str] | None = None
    ):
        """
        Rename the assignee of the call to the given index.
        Optionally, you can pass a modifier that receives the value of the parameter and returns a new name.
        """

        if call.assignee is None:
            print(f"Call {call} has no assignee")
            return
        if index >= len(call.params):
            print(f"Call {call} has no parameter at index {index}")
            return
        param = call.params[index]
        if param is None:
            print(f"Call {call} param at index {index} is not const: {call.params_op[index].dstr()}")
            return
        if modifier is None and not isinstance(param, str):
            print(f"Call {call} param at index {index} is not a string: {param}")
            return

        name: str = modifier(param) if modifier else param
        self.__rename_mop(call.assignee, name, modifications)

    # noinspection PyMethodMayBeStatic
    def __retype_mop(self, op: mop_t, typ: tinfo_t, modifications: Modifications):
        """Try to retype the mop to the given name"""
        if op.t == ida_hexrays.mop_v:
            # Global variable
            modifications.modify_global(op.g, VariableModification(type=typ))
        elif op.t == ida_hexrays.mop_l:
            if op.l.off != 0:
                # Local variable with offset are unsupported for now
                print(f"Could not retype mop {op.dstr()} to {typ}: has offset {op.l.off}")
            # Local variable
            modifications.modify_local(op.l.var().name, VariableModification(type=typ))
        elif op.t == ida_hexrays.mop_a:
            # Deref the type
            self.__retype_mop(op.a, typ.get_pointed_object(), modifications)
        elif op.t == ida_hexrays.mop_d:
            inner_insn: minsn_t = op.d
            if ida_hexrays.is_mcode_xdsu(inner_insn.opcode):
                self.__retype_mop(inner_insn.l, typ, modifications)
            else:
                # TODO support fields
                print(f"Could not retype mop {op.dstr()} to {typ}: unsupported type {op.t} yet")

    # noinspection PyMethodMayBeStatic
    def __rename_mop(self, op: mop_t, name: str, modifications: Modifications, follow_address: bool = True):
        """Try to rename the mop to the given name"""
        if op.t == ida_hexrays.mop_v:
            # Global variable
            modifications.modify_global(op.g, VariableModification(name=name, force_name_change=False))
        elif op.t == ida_hexrays.mop_l:
            if op.l.off != 0:
                # Local variable with offset are unsupported for now
                print(f"Could not rename mop {op.dstr()} to {name}: has offset {op.l.off}")
            # Local variable
            modifications.modify_local(op.l.var().name, VariableModification(name=name, force_name_change=False))
        elif follow_address and op.t == ida_hexrays.mop_a:
            # Variable whose address is taken
            self.__rename_mop(op.a, name, modifications, follow_address)
        elif op.t == ida_hexrays.mop_d:
            inner_insn: minsn_t = op.d
            if ida_hexrays.is_mcode_xdsu(inner_insn.opcode):
                self.__rename_mop(inner_insn.l, name, modifications, follow_address)
            else:
                # TODO support fields
                print(f"Could not rename mop {op.dstr()} to {name}: unsupported type {op.t} yet")


class FuncHandlerByNameWithStringFinder(FuncHandler, ABC):
    def __init__(self, name: str, func_type: tinfo_t | None, search_string: str, is_call: bool):
        super().__init__(name)
        self.search_string: str = search_string
        self.func_type: tinfo_t | None = func_type
        self.is_call = is_call

    def get_source_xref(self) -> SourceXref | None:
        if self.func_type is None:
            return None

        existing = memory.ea_from_name(self.name)
        if existing is not None:
            return FuncXref(existing)

        if self.is_call:
            searched = xrefs.find_static_caller_for_string(self.search_string)
        else:
            searched = xrefs.find_func_containing_string(self.search_string)
        if searched is None:
            print("Could not find function", self.name)
            # Could not find the function
            return None

        memory.set_name(searched, self.name, retry=True)
        print(f"Found {self.name} at {searched:#x}, changing name")
        if not tif.apply_tinfo_to_ea(self.func_type, searched):
            print(f"Could not apply type {self.func_type} to {searched:#x}")
            return None

        return FuncXref(searched)


class FuncHandlerVirtualGetter(FuncHandler):
    def __init__(
        self, name: str, obj_type: tinfo_t | None, offset: int, name_index: int, rename_prefix: str | None = None
    ):
        super().__init__(name)
        self.obj_type: tinfo_t | None = obj_type
        self.offset = offset
        self.name_index = name_index
        self.rename_modifier = lambda n: f"{rename_prefix}{n}" if rename_prefix is not None else None

    def get_source_xref(self) -> SourceXref | None:
        if self.obj_type is None:
            return None
        return IndirectCallXref(self.obj_type, self.offset)

    def on_call(self, call: Call, modifications: Modifications):
        self._rename_assignee_by_index(modifications, call, self.name_index, self.rename_modifier)


class FuncHandlerVirtualSetter(FuncHandler):
    def __init__(
        self,
        name: str,
        obj_type: tinfo_t | None,
        offset: int,
        name_index: int,
        rename_index: int,
        rename_prefix: str | None = None,
    ):
        super().__init__(name)
        self.obj_type: tinfo_t | None = obj_type
        self.offset = offset
        self.name_index = name_index
        self.rename_index = rename_index
        self.rename_modifier = lambda n: f"{rename_prefix}{n}" if rename_prefix is not None else None

    def get_source_xref(self) -> SourceXref | None:
        if self.obj_type is None:
            return None
        return IndirectCallXref(self.obj_type, self.offset)

    def on_call(self, call: Call, modifications: Modifications):
        self._rename_parameter_by_index(modifications, call, self.name_index, self.rename_index, self.rename_modifier)



__all__ = ["objc_selector_hexrays_hooks_t"]

import re

import ida_hexrays
from ida_hexrays import Hexrays_Hooks, carg_t, carglist_t, cexpr_t, cfunc_t, citem_t
from ida_kernwin import simpleline_t
from ida_lines import (
    COLOR_OFF,
    COLOR_ON,
    COLSTR,
    SCOLOR_ADDR,
    SCOLOR_DEMNAME,
    SCOLOR_IMPNAME,
    SCOLOR_LOCNAME,
    SCOLOR_SYMBOL,
    tag_remove,
)
from ida_pro import strvec_t
from idahelper import objc
from idahelper.ast import cexpr

SELECTOR_MARKER = "!@#$sel$#@!"
COMMA_COLORED = COLSTR(",", SCOLOR_SYMBOL)
INSIGNIFICANT_LENGTH_FOR_LINE = 5
MAX_LINE_SIZE = 120

SEL_TOKEN_REGEX = re.compile(
    "(?P<prefix>"
    + re.escape(COMMA_COLORED + " ")
    + r")?"
    + re.escape(COLOR_ON + SCOLOR_ADDR)
    + r"(?P<index>[0-9A-Fa-f]{16})"
    + re.escape(COLOR_ON + SCOLOR_ADDR)
    + r"(?P=index)"
    + re.escape(COLSTR('"', SCOLOR_SYMBOL))
    + re.escape(COLOR_ON + SCOLOR_LOCNAME)
    + r"(?P<selector>[A-Za-z0-9_:]+)"
    + '"'
    + re.escape(COLOR_OFF + SCOLOR_LOCNAME)
)
"""
a regex for a possible selector in IDA's pseudocode, with support for the prefix ", "
Its groups are: prefix, index, and selector.
"""

# noinspection RegExpDuplicateCharacterInClass
CLASS_TOKEN_REGEX = re.compile(
    re.escape(COLOR_ON + SCOLOR_ADDR)
    + r"(?P<index>[0-9A-Fa-f]{16})"
    + re.escape(COLSTR("&", SCOLOR_SYMBOL))
    + re.escape(COLOR_ON + SCOLOR_ADDR)
    + r"(?P<index2>[0-9A-Fa-f]{16})"
    + re.escape(COLOR_ON)
    + ("[" + re.escape(SCOLOR_IMPNAME) + "|" + re.escape(SCOLOR_DEMNAME) + "]")
    + "OBJC_CLASS___"
    + "(?P<class>[A-Za-z0-9_]+)"
    + re.escape(COLOR_OFF)
    + ("[" + re.escape(SCOLOR_IMPNAME) + "|" + re.escape(SCOLOR_DEMNAME) + "]")
    + "(?P<postfix>"
    + re.escape(COMMA_COLORED)
    + r" ?)?"
)
"""
a regex for possible obj-c class in IDA's pseudocode, with support for the postfix ","
Its groups are: index, index2, class and postfix.
"""


class objc_selector_hexrays_hooks_t(Hexrays_Hooks):
    def func_printed(self, cfunc: cfunc_t) -> int:  # noqa: C901
        selectors_to_remove: dict[int, str] = {}  # obj_id -> selector
        classes_to_remove: set[int] = set()  # obj_id
        index_to_sel: dict[int, str] = {}  # index, selector
        class_indices_to_remove: set[int] = set()  # index

        for i, call_item in enumerate(cfunc.treeitems):
            call_item: citem_t
            # Get the index of the selector/class AST element
            if call_item.obj_id in selectors_to_remove:
                index_to_sel[i] = selectors_to_remove.pop(call_item.obj_id)
            elif call_item.obj_id in classes_to_remove:
                class_indices_to_remove.add(i)
                classes_to_remove.remove(call_item.obj_id)

            elif call_item.op == ida_hexrays.cot_call:
                call_expr: cexpr_t = call_item.cexpr

                # 1. Check if the function name looks like an Obj-C method
                call_func_name = cexpr.get_call_name(call_expr)
                if call_func_name is None or not objc.is_objc_method(call_func_name):
                    continue

                # 2. Collect selector from arglist
                arglist: carglist_t = call_expr.a
                if len(arglist) < 2:
                    print("[Error]: Obj-C method call with less than 2 arguments:", call_expr.dstr())
                    continue
                sel_arg: carg_t = arglist[1]
                if sel_arg.op != ida_hexrays.cot_str:
                    print("[Error]: Obj-C method call with non-string selector:", call_expr.dstr())
                    continue
                selectors_to_remove[sel_arg.obj_id] = sel_arg.string

                # 3. Check if the function is a class method
                if objc.is_objc_static_method(call_func_name):
                    # 4. Check if the class name is a ref to obj
                    class_arg: carg_t = arglist[0]
                    if class_arg.op != ida_hexrays.cot_ref or class_arg.x.op != ida_hexrays.cot_obj:
                        print("[Error]: Obj-C class method with unsupported class", call_expr.dstr())
                        continue
                    # 5. Collect the class name
                    classes_to_remove.add(class_arg.obj_id)

        if selectors_to_remove or classes_to_remove:
            print("[Error]: unmatched Obj-C selectors in the function: ", hex(cfunc.entry_ea))
        elif index_to_sel:
            modify_text(cfunc, index_to_sel, class_indices_to_remove)
        return 0


def modify_text(cfunc: cfunc_t, index_to_sel: dict[int, str], class_indices_to_remove: set[int]):
    # Early return if no tokens to replace
    if not index_to_sel:
        return

    ps: strvec_t = cfunc.get_pseudocode()
    lines_marked_for_removal: list[simpleline_t] = []
    for i, line in enumerate(ps):
        line: simpleline_t
        prev_line = ps[i - 1].line if i != 0 else ""
        should_merge = modify_selectors(index_to_sel, line, prev_line)
        should_merge |= modify_class(class_indices_to_remove, line, prev_line)

        if should_merge and i != 0:
            lines_marked_for_removal.append(line)
            merge(line.line, ps[i - 1])

    # Remove lines that are marked for removal
    for line_to_remove in reversed(lines_marked_for_removal):
        ps.erase(line_to_remove)


def modify_selectors(index_to_sel: dict[int, str], line: simpleline_t, prev_line: str):
    """Try to remove selectors from a line. Returns whether we should merge the line with the previous line"""
    should_merge = False
    # Reverse the results so indices will not change
    for result in reversed(list(re.finditer(SEL_TOKEN_REGEX, line.line))):
        result: re.Match
        index = int(result.group("index"), 16)
        if index in index_to_sel:
            # We found a selector token, remove it from the list
            sel = index_to_sel.pop(index)
            if sel != result.group("selector"):
                print("[Error]: selector mismatch. Expected:", sel, "Actual:", result.group("selector"))
                continue

            # Remove the selector, check if we need to merge lines
            left, right = result.span()
            before_selector, after_selector = line.line[:left], line.line[right:]
            should_merge = should_merge or should_merge_line(before_selector, after_selector, prev_line)
            line.line = before_selector + after_selector
    return should_merge


def modify_class(class_indices_to_remove: set[int], line: simpleline_t, prev_line: str):
    """Try to remove class from a line. Returns whether we should merge the line with the previous line"""
    should_merge = False

    # Reverse the results so indices will not change
    for result in reversed(list(re.finditer(CLASS_TOKEN_REGEX, line.line))):
        result: re.Match
        index = int(result.group("index"), 16)
        index2 = int(result.group("index2"), 16)
        if index in class_indices_to_remove:
            # We found a class token, remove it from the list
            class_indices_to_remove.remove(index)

            if index2 != index + 1:
                print("[Error]: class indices mismatch for second object. Expected:", index + 1, "Actual:", index2)
                continue

            # Remove the class, check if we need to merge lines
            left, right = result.span()
            before_class, after_class = line.line[:left], line.line[right:]
            should_merge = should_merge or should_merge_line(before_class, after_class, prev_line)
            line.line = before_class + after_class
    return should_merge


def should_merge_line(before_selector: str, after_selector: str, prev_line: str) -> bool:
    """
    Given a `line` with a selector marker and the previous line, should we merge the two lines.
    """
    before_without_tags = tag_remove(before_selector)

    # We only remove lines that starts with the selector
    if before_without_tags and not before_without_tags.isspace():
        return False

    after_without_tags = tag_remove(after_selector)
    # If the line is short, allow merging
    if len(after_without_tags) < INSIGNIFICANT_LENGTH_FOR_LINE:
        return True

    # Merge if it will not lead to a long line
    prev_line_without_tags = tag_remove(prev_line)
    return len(after_without_tags) + len(prev_line_without_tags) < MAX_LINE_SIZE


def merge(text: str, line: simpleline_t) -> None:
    """Merge `text` to the end of `line`. If `line` ends with a comma, the comma will be removed."""
    line_without_tags = tag_remove(line.line)
    if line_without_tags.endswith(","):
        last_comma_index = line.line.rfind(COMMA_COLORED)
        line.line = line.line[:last_comma_index] + line.line[last_comma_index + len(COMMA_COLORED) :]
    line.line += text.strip()


def to_hex(n: int, *, length: int) -> str:
    """Convert an integer to a hex string with leading zeros"""
    return f"{n:0{length}X}"



__all__ = ["run_callback"]

from collections.abc import Callable

import idaapi
from ida_funcs import func_t
from idahelper import file_format

from ioshelper.base.reloadable_plugin import PluginCore

from ..kernelcache.func_renamers import apply_global_rename, apply_pac
from ..kernelcache.kalloc_type import apply_kalloc_types
from .clang_blocks import run_objc_plugin_on_func, try_add_block_arg_byref_to_func
from .outline import mark_all_outline_functions

RUN_GLOBAL_ANALYSIS = 1
RUN_LOCAL_ANALYSIS = 2
NETNODE_NAME = "$ idaioshelper"


def run_callback(_core: PluginCore) -> Callable[[int], None]:
    def run(value: int):
        # Here you can implement the logic that uses the core and the value
        print(f"[Debug] iOS helper run({value})")
        if value == RUN_GLOBAL_ANALYSIS:
            run_global_analysis()
        elif value == RUN_LOCAL_ANALYSIS:
            ea = read_ea_arg()
            if ea is None:
                print("[Error] No function address provided for local analysis.")
                return
            func = idaapi.get_func(ea)
            if func is None:
                print(f"[Error] No function found at address {ea:X}.")
                return

            run_local_analysis(func)

    return run


def run_global_analysis():
    print("[Info] Running global analysis...")

    print("[Info] Running outline detection...")
    mark_all_outline_functions()
    print("[Info] Outline detection completed.")

    if file_format.is_kernelcache():
        print("[Info] Applying kalloc types...")
        apply_kalloc_types()
        print("[Info] Apply kalloc types completed.")

        print("[Info] Running global renaming...")
        apply_global_rename()
        print("[Info] Global renaming completed.")

    print("[Info] Global analysis completed.")


def run_local_analysis(func: func_t):
    print("[Info] Running local analysis...")
    # Implement local analysis logic here
    print("[Info] Use builtin Obj-C plugin to restore blocks")
    run_objc_plugin_on_func(func.start_ea)
    print("[Info] Use builtin Obj-C plugin to restore blocks completed.")

    print("[Info] Try restore byref arguments in blocks")
    try_add_block_arg_byref_to_func(func.start_ea)
    print("[Info] Try restore byref arguments in blocks completed.")

    print("[Info] Try use PAC to apply types to local variables and fields")
    apply_pac(func)
    print("[Info] Try use PAC to apply types to local variables and fields completed.")

    print("[Info] Local analysis completed.")


def write_ea_arg(ea: int):
    n = idaapi.netnode()
    n.create(NETNODE_NAME)
    n.altset(1, ea, "R")


def read_ea_arg() -> int | None:
    n = idaapi.netnode(NETNODE_NAME)
    val = n.altval(1, "R")
    n.kill()
    return val if val != 0 else None



import ida_kernwin
import idaapi
from ida_kernwin import action_handler_t

from ioshelper.base.reloadable_plugin import UIAction, UIActionsComponent

from .cpp_vtbl import get_vtable_call, show_vtable_xrefs

ACTION_ID = "ioshelper:jump_to_vtbl_xrefs"

jump_to_vtable_component = UIActionsComponent.factory(
    "Jump to VTables xrefs",
    [
        lambda core: UIAction(
            ACTION_ID,
            idaapi.action_desc_t(
                ACTION_ID,
                "Jump to VTables xrefs",
                JumpToVtablesXrefs(),
                "Shift+X",
            ),
            dynamic_menu_add=lambda widget, popup: idaapi.get_widget_type(widget) == idaapi.BWN_PSEUDOCODE
            and get_vtable_call() is not None,
        )
    ],
)


class JumpToVtablesXrefs(action_handler_t):
    def activate(self, ctx: ida_kernwin.action_ctx_base_t):
        show_vtable_xrefs()
        return False

    def update(self, ctx):
        return idaapi.AST_ENABLE_ALWAYS



__all__ = ["show_segment_xrefs_component"]

import ida_kernwin
import idaapi
from ida_kernwin import action_handler_t

from ioshelper.base.reloadable_plugin import UIAction, UIActionsComponent

from .segment_xrefs import can_show_segment_xrefs, get_current_expr, show_segment_xrefs

ACTION_ID = "ioshelper:show_segment_xrefs"

show_segment_xrefs_component = UIActionsComponent.factory(
    "Show Xrefs inside segment",
    [
        lambda core: UIAction(
            ACTION_ID,
            idaapi.action_desc_t(
                ACTION_ID,
                "Show Xrefs inside segment",
                ShowSegmentXrefsAction(),
                "Ctrl+Shift+X",
            ),
            dynamic_menu_add=lambda widget, popup: can_show_segment_xrefs(widget),
        )
    ],
)


class ShowSegmentXrefsAction(action_handler_t):
    def activate(self, ctx: ida_kernwin.action_ctx_base_t):
        expr = get_current_expr(ctx.widget)
        if expr is None:
            print("[Error] No expression found in the current context.")
            return False
        show_segment_xrefs(expr, func_ea=ctx.cur_func.start_ea)
        return False

    def update(self, ctx):
        return idaapi.AST_ENABLE_ALWAYS



__all__ = ["try_add_block_arg_byref_to_func"]

from collections.abc import Callable
from dataclasses import dataclass

import ida_hexrays
from ida_hexrays import (
    cexpr_t,
    mba_t,
    minsn_t,
    mop_t,
)
from ida_typeinf import tinfo_t
from idahelper import tif
from idahelper.ast import cexpr, cfunc, lvars
from idahelper.ast.lvars import VariableModification
from idahelper.microcode import mba, mblock, mop

from .block import (
    FLAG_BLOCK_HAS_COPY_DISPOSE,
    block_member_is_arg_field,
    get_ida_block_lvars,
)
from .block_arg_byref import (
    BlockArgByRefField,
    create_block_arg_byref_type,
    is_block_arg_byref_type,
)
from .utils import StructFieldAssignment, get_struct_fields_assignments


def try_add_block_arg_byref_to_func(ea: int) -> bool:
    func = cfunc.from_ea(ea)
    if func is None:
        print(f"Failed to decompile func at {ea:X}")
        return False

    block_lvars = get_ida_block_lvars(func)
    if not block_lvars:
        return False

    # Scan the cfunc for possible ref args for blocks
    assignments = get_struct_fields_assignments(func, block_lvars)
    stack_off_to_its_assignment: dict[int, StructFieldAssignment] = {}  # stack_offset -> assignment
    for lvar in block_lvars:
        if lvar.name not in assignments:
            print(f"[Error] Block variable {lvar.name} has no assignments")
            continue
        stack_off_to_its_assignment.update(get_by_ref_args_for_block_candidates(assignments[lvar.name]))

    if not stack_off_to_its_assignment:
        return False

    # scan the microcode using the offsets to see if any of them is a start of a by ref arg struct.
    lvar_modifications: dict[str, VariableModification] = {}  # lvar_name -> type_modification
    for result in ScanForRefArg(set(stack_off_to_its_assignment.keys())).scan(func.mba):
        new_type = create_block_arg_byref_type(result.initialization_ea, result.variable.size, result.has_helpers)
        block_arg_by_ref_lvar = cfunc.get_lvar_by_offset(func, result.initial_stack_offset)
        lvar_modifications[block_arg_by_ref_lvar.name] = VariableModification(type=new_type)

        # given: a.b = &block_by_ref, set b's type to block_by_ref*
        # It would not be a cast assign, as the type is already a pointer.
        assignment = stack_off_to_its_assignment[result.initial_stack_offset]
        set_new_type_for_member(assignment, tif.pointer_of(new_type))

    if lvar_modifications:
        # Apply new types for the lvars
        if not lvars.perform_lvar_modifications_by_ea(func.entry_ea, lvar_modifications):
            print("[Error] Failed to modify lvars")

        return True
    return False


def set_new_type_for_member(assignment: StructFieldAssignment, new_type: tinfo_t) -> bool:
    """Set the new type for the member of the struct"""
    if not tif.set_udm_type(assignment.type, assignment.member, new_type):
        print(f"[Error] Failed to set udm type for {assignment.member.name} in {assignment.type.get_type_name()}")
        return False
    return True


@dataclass
class ScanForBlockArgByRefState:
    current_field: BlockArgByRefField
    initialization_ea: int | None
    has_helpers: bool | None
    initial_stack_offset: int | None
    isa: mop_t | None
    flags: mop_t | None
    size: mop_t | None
    helper_keep: mop_t | None
    helper_dispose: mop_t | None
    variable: mop_t | None

    @staticmethod
    def initial() -> "ScanForBlockArgByRefState":
        return ScanForBlockArgByRefState(
            current_field=BlockArgByRefField.ISA,
            initialization_ea=None,
            has_helpers=None,
            initial_stack_offset=None,
            isa=None,
            flags=None,
            size=None,
            helper_keep=None,
            helper_dispose=None,
            variable=None,
        )

    def expected_offset(self) -> int:
        """Returns the expected relative offset for the next field to read"""
        assert self.initial_stack_offset is not None, "stack_offset must be set before calling expected_offset"

        return self.initial_stack_offset + self.current_field.get_offset(self.has_helpers)


class ScanForRefArg:
    def __init__(self, possible_stack_offsets: set[int]):
        self._state = ScanForBlockArgByRefState.initial()
        self._possible_stack_offsets = possible_stack_offsets
        self._results: list[ScanForBlockArgByRefState] = []
        self._state_handlers: dict[BlockArgByRefField, Callable[[mop_t, int], None]] = {
            BlockArgByRefField.ISA: self._isa,
            BlockArgByRefField.FORWARDING: self._forwarding,
            BlockArgByRefField.FLAGS: self._flags,
            BlockArgByRefField.SIZE: self._size,
            BlockArgByRefField.HELPER_KEEP: self._helper_keep,
            BlockArgByRefField.HELPER_DISPOSE: self._helper_dispose,
            BlockArgByRefField.VARIABLE: self._variable,
        }

    def scan(self, func_mba: mba_t) -> list[ScanForBlockArgByRefState]:
        # Clean state
        self._state = ScanForBlockArgByRefState.initial()
        self._results = []

        for block in mba.blocks(func_mba):
            for insn in mblock.instructions(block):
                self._on_instruction(insn)
        return self._results

    def _on_instruction(self, insn: minsn_t) -> None:
        # initialization of the block by reg arg is done using mov only.
        if insn.opcode != ida_hexrays.m_mov:
            self._state = ScanForBlockArgByRefState.initial()
            return

        # Check if the instruction is a mov to a stack variable
        offset = mop.get_stack_offset(insn.d)
        if offset is None:
            self._state = ScanForBlockArgByRefState.initial()
            return

        # Check if the write offset is correct
        if self._state.initial_stack_offset is not None and offset != self._state.expected_offset():
            self._state = ScanForBlockArgByRefState.initial()
            return

        if self._state.initialization_ea is None:
            self._state.initialization_ea = insn.ea

        self._state_handlers[self._state.current_field](insn.l, offset)

    def _isa(self, value: mop_t, offset: int) -> None:
        if (
            offset not in self._possible_stack_offsets
            or value.size != 8
            or value.t != ida_hexrays.mop_n
            or value.unsigned_value() != 0
        ):
            self._state = ScanForBlockArgByRefState.initial()
            return

        self._state.current_field = BlockArgByRefField.FORWARDING
        self._state.initial_stack_offset = offset
        self._state.isa = value

    def _forwarding(self, value: mop_t, _offset: int) -> None:
        if value.size != 8 or mop.get_stack_offset(value) != self._state.initial_stack_offset:
            self._state = ScanForBlockArgByRefState.initial()
            return

        self._state.current_field = BlockArgByRefField.FLAGS
        self._state.forwarding = value

    def _flags(self, value: mop_t, _offset: int) -> None:
        # Here we can encounter either a DWORD assignment for flags,
        # or a QWORD assignment for both the flags and size.
        if value.size not in (4, 8):
            self._state = ScanForBlockArgByRefState.initial()
            return

        numeric_value = mop.get_const_int(value, is_signed=False)
        if numeric_value is None:
            self._state = ScanForBlockArgByRefState.initial()
            return

        if value.size == 4:
            self._state.current_field = BlockArgByRefField.SIZE
            self._state.flags = value
            flags = numeric_value
        else:
            flags = numeric_value & 0xFF_FF_FF_FF
            size = (numeric_value >> 32) & 0xFFFF
            # Create new mop for size and flags
            self._state.flags = mop.from_const_value(flags, 4)
            self._state.size = mop.from_const_value(size, 4)

            if flags & FLAG_BLOCK_HAS_COPY_DISPOSE:
                self._state.has_helpers = True
                self._state.current_field = BlockArgByRefField.HELPER_KEEP
            else:
                self._state.has_helpers = False
                self._state.current_field = BlockArgByRefField.VARIABLE

    def _size(self, value: mop_t, _offset: int) -> None:
        if value.size != 4:
            self._state = ScanForBlockArgByRefState.initial()
            return

        # Check that value is a constant
        numeric_value = mop.get_const_int(value, is_signed=False)
        if numeric_value is None:
            self._state = ScanForBlockArgByRefState.initial()
            return

        self._state.size = value

        # Flags cannot be none because we checked it before
        flags = mop.get_const_int(self._state.flags, is_signed=False)
        if flags & FLAG_BLOCK_HAS_COPY_DISPOSE:
            self._state.has_helpers = True
            self._state.current_field = BlockArgByRefField.HELPER_KEEP
        else:
            self._state.has_helpers = False
            self._state.current_field = BlockArgByRefField.VARIABLE

    def _helper_keep(self, value: mop_t, _offset: int) -> None:
        if value.size != 8 or not value.is_glbaddr():
            self._state = ScanForBlockArgByRefState.initial()
            return

        self._state.helper_keep = value
        self._state.current_field = BlockArgByRefField.HELPER_DISPOSE

    def _helper_dispose(self, value: mop_t, _offset: int) -> None:
        if value.size != 8 or not value.is_glbaddr():
            self._state = ScanForBlockArgByRefState.initial()
            return

        self._state.helper_keep = value
        self._state.current_field = BlockArgByRefField.VARIABLE

    def _variable(self, value: mop_t, _offset: int) -> None:
        self._state.variable = value
        self._results.append(self._state)

        # Reset the state
        self._state = ScanForBlockArgByRefState.initial()


def get_by_ref_args_for_block_candidates(assignments: list[StructFieldAssignment]) -> dict[int, StructFieldAssignment]:
    """
    Filter assignments such that the rvalue is ref to stack variable.
    Return mapping of var_stack_offset to assignment
    """
    possible_stack_offsets = {}
    for assignment in assignments:
        stack_offset = -1
        # Skip fields that are not args
        if not block_member_is_arg_field(assignment.member):
            continue
        # Find assignments that are refs to the stack: "block.lvar2 = &v8" or, "block.lvar2 = v8" (and v8 is array type)
        expr = cexpr.strip_casts(assignment.expr)
        # block.lvar2 = &v8
        if expr.op == ida_hexrays.cot_ref:
            refed_expr: cexpr_t = expr.x
            if refed_expr.op != ida_hexrays.cot_var:
                continue
            # Check if the variable is not already handled
            if is_block_arg_byref_type(refed_expr.v.getv().type()):
                continue
            # Return the stack offset of the variable
            stack_offset = refed_expr.v.getv().get_stkoff()
        # block.lvar2 = v8 (and v8 is array type)
        elif expr.op == ida_hexrays.cot_var:
            # The variable could not have been handled, as it should be a ref in this case
            # Check it is an array type, as this is the only situation I can think of that
            # will represent a byref arg block and will not be a ref.
            if not expr.v.getv().type().is_array():
                continue
            stack_offset = expr.v.getv().get_stkoff()

        if stack_offset != -1:
            possible_stack_offsets[stack_offset] = assignment

    return possible_stack_offsets



__all__ = ["component"]

from ioshelper.base.reloadable_plugin import OptimizersComponent

from .error_case_optimizer import log_error_case_optimizer_t
from .log_enabled_optimizer import os_log_enabled_optimizer_t
from .log_macro_optimizer import optimizer as log_macro_optimizer

component = OptimizersComponent.factory(
    "os_log optimizer", [log_error_case_optimizer_t, log_macro_optimizer, os_log_enabled_optimizer_t]
)
